#include "Cpf.h"
#include <fstream>
#include <sstream>
#include <math.h>

/******************************************************************************
Constructor 
******************************************************************************/
Cpf::Cpf()
{
    m_elemType = Element::Type::cpf;
    m_xPositionTx = m_yPositionTx = m_zPositionTx = 0.0;  
    m_xPositionRx = m_yPositionRx = m_zPositionRx = 0.0;  
    m_xVelocityTx = m_yVelocityTx = m_zVelocityTx = 0.0; 
    m_xVelocityRx = m_yVelocityRx = m_zVelocityRx = 0.0; 
    m_xAberrationTx = m_yAberrationTx = m_zAberrationTx = 0.0;  
    m_xAberrationRx = m_yAberrationRx = m_zAberrationRx = 0.0;  
    m_relCorrTx = m_relCorrRx = 0.0;
    m_biDirectionalVel = m_biDirectionalCor = m_biDirectionalPos = false;
    m_complete =  m_irvCompatible = false;
    m_lightTimeCorrected = true;
    m_referenceFrame = unknownRefFrame;
    m_targetType = unknownTargetType;
    m_angleType = RotationAngles::Type::unknownAngleType;
    m_nlr = 0;

};

/******************************************************************************
Destructor
******************************************************************************/
Cpf::~Cpf()
{
}

/******************************************************************************
Determine enumerated record type
******************************************************************************/
Cpf::RecordType Cpf::recordType( const std::string& s)
{
    Cpf::RecordType type = Cpf::RecordType::unknownRecordType;
    if( s == "h1" || s == "H1" ) type = Cpf::RecordType::headerBasic1;
    if( s == "h2" || s == "H2" ) type = Cpf::RecordType::headerBasic2;
    if( s == "h3" || s == "H3" ) type = Cpf::RecordType::headerAccuracy;
    if( s == "h4" || s == "H4" ) type = Cpf::RecordType::headerTransponder;
    if( s == "h5" || s == "H5" ) type = Cpf::RecordType::headerCoM;
    if( s == "h9" || s == "H9" ) type = Cpf::RecordType::lastHeader;
    if( s == "10" ) type = Cpf::RecordType::recordPosition;
    if( s == "20" ) type = Cpf::RecordType::recordVelocity;
    if( s == "30" ) type = Cpf::RecordType::recordCorrection;
    if( s == "40" ) type = Cpf::RecordType::recordTransponder;
    if( s == "50" ) type = Cpf::RecordType::recordOffset;
    if( s == "60" ) type = Cpf::RecordType::recordAngle;
    if( s == "70" ) type = Cpf::RecordType::recordEOP;
    if( s == "99" ) type = Cpf::RecordType::lastRecord;
    return type;    
}


/******************************************************************************
Header type 1 Basic information (required)
 1-2 A2 Record Type (= "H1")
 4-6 A3 "CPF"
 8-9 I2 Format Version
12-14 A3 Ephemeris Source (e.g., "HTS ", "UTX ")
16-19 I4 Year of ephemeris production
21-22 I2 Month of ephemeris production
24-25 I2 Day of ephermeris production
27-28 I2 Hour of ephemeris production (UTC)
31-34 I4 Ephemeris Sequence number
36-45 A10 Target Name
47-56 A10 Notes (e.g., "041202","DE-403")
******************************************************************************/
void Cpf::parseBasicHeader1( const std::string& line, Header* header )
{
    yourASSERTION( header );

    if( header->nh1++ > 1 )
        yourTHROW( "Multiple H1 records encountered.");

    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    std::string format  = convert<std::string>( s, ok ); 
    std::string version = convert<std::string>( s, ok ); 
    std::transform(recType.begin(), recType.end(), recType.begin(), (int(*)(int)) toupper);
    std::transform(format.begin(), format.end(), format.begin(), (int(*)(int)) toupper);
    if( recType != "H1" || format != "CPF" )
        yourTHROW("Error reading cpf record H1.");

    header->source   = convert<std::string>( s, ok ); 
    long  year       = convert<long>(s, ok); 
    long  month      = convert<long>(s, ok); 
    long  day        = convert<long>(s, ok); 
    long  hour       = convert<long>(s, ok); 
    header->sequence = convert<long>(s, ok);

    if( !ok )
        yourTHROW("Error reading cpf record H1.");

    // Target name may or may not exist. Ignore the ok flag.
    header->target   = convert<std::string>(s, ok);

    // Notes may or may not exist. Ignore the ok flag.
    header->notes    = convert<std::string>(s, ok);

    YourTimeDateClass date;
    date.setDate( year, month, day, hour, 0, 0, 0, 0 );
    header->created  = date;
    return;
}

/******************************************************************************
Header type 2 Basic information - 2 (required)
 1-2  A2  Record Type (= "H2")
 4-1  1I8 COSPAR ID
13-1  6I4 SIC
18-2  5I8 NORAD ID
27-30 I4 Starting Year
32-33 I2 Starting Month
35-36 I2 Staring Day
38-39 I2 Starting Hour (UTC)
41-42 I2 Starting Minute (UTC)
44-45 I2 Starting Second (UTC)
47-50 I4 Ending Year
52-53 I2 Ending Month
55-56 I2 Ending Day
58-59 I2 Ending Hour (UTC)
61-62 I2 Ending Minute (UTC)
64-65 I2 Ending Second (UTC)
67-71 I5 Time between table entries (UTC seconds)(=0 if variable)
73 I1 Compatibility with TIVs = 1 (=> integratable, geocentric ephemeris)
75 I1 Target type
  1 = passive artificial satellite
  2 = passive lunar reflector
  3 = synchronous transponder
  4 = asynchronous transponder
77-78 I2 Reference frame
  0 = geocentric true body fixed (default)
  1 = geocentric space fixed (i.e. Inertial) (True of Date)
  2 = geocentric space fixed (Mean of Date J2000)
80 I1 Rotational angle type
  0 = Not Applicable
  1 = Lunar Euler angles: ?, ?, and ?
  2 = North pole Right Ascension and Declination, and angle to prime meridian (?0, ?0, and W)
82 I1 Centre of mass correction
  0 = none applied, predcition is for centre of mass of target
  1 = applied, prediction is for retro-reflector array
******************************************************************************/
void Cpf::parseBasicHeader2( const std::string& line, Header* header )
{
    yourASSERTION( header );
    if( header->nh2++ > 1 )
        yourTHROW( "Multiple H2 records encountered.");

    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "h2" && recType != "H2" )
        yourTHROW( "Error reading cpf record H2.");

    header->cospar = convert<std::string>( s, ok ); 
    header->sic    = convert<std::string>( s, ok ); 
    header->norad  = convert<std::string>( s, ok ); 

    // Extract dates
    long year1, month1, day1, hour1, minute1, second1;
    long year2, month2, day2, hour2, minute2, second2;
    s  >> year1 >> month1 >> day1 >> hour1 >> minute1 >> second1 
       >> year2 >> month2 >> day2 >> hour2 >> minute2 >> second2; 
    YourTimeDateClass epoch;
    epoch.setDate( year1, month1, day1, hour1, minute1, second1, 0, 0 );
    header->startDate = epoch;
    epoch.setDate(   year2, month2, day2, hour2, minute2, second2, 0, 0 );
    header->endDate = epoch;

    header->interval  = convert<long>( s, ok ); 
    long tivComp      = convert<long>( s, ok ); 
    long targetType   = convert<long>( s, ok ); 
    long refFrame     = convert<long>( s, ok ); 
    long angleType    = convert<long>( s, ok ); 
    if( !ok )
        yourTHROW( "Error reading cpf record H2.");

    header->irvCompatible = ( tivComp == 1 );
    long comApplied   = convert<long>( s, ok ); if( ok ) header->comApplied = ( comApplied == 1 );

    // Translate and set target type enumeration
    switch( targetType )
    {
        case 1 : header->targetType = Cpf::TargetType::passiveSatellite;  break;
        case 2 : header->targetType = Cpf::TargetType::passiveLunar;      break;
        case 3 : header->targetType = Cpf::TargetType::synchTransponder;  break;
        case 4 : header->targetType = Cpf::TargetType::asynchTransponder; break;
        default:
        case 0 : header->targetType = Cpf::TargetType::unknownTargetType; break;
    }

    // Translate and set reference frame enumeration
    switch( refFrame )
    {
        case 0 : header->referenceFrame = Cpf::ReferenceFrame::itrf;            break;
        case 1 : header->referenceFrame = Cpf::ReferenceFrame::tod;             break;
        case 2 : header->referenceFrame = Cpf::ReferenceFrame::mod;             break;
        default: header->referenceFrame = Cpf::ReferenceFrame::unknownRefFrame; break;
    }

    // Translate and set rotation angle type enumeration
    switch( angleType )
    {
        case 1 : header->angleType = RotationAngles::Type::lunarEular;       break;
        case 2 : header->angleType = RotationAngles::Type::raDecMeridian;    break;
        default:
        case 0 : header->angleType = RotationAngles::Type::unknownAngleType; break;
    }

    s.str( header->sic );
    ok = true;
    header->identifier = convert<long>( s, ok );
    if( !ok )
        yourTHROW( "Error reading cpf record H2.");

    return;
}

/******************************************************************************
Header type 3 Expected accuracy 
1-2   A2 Record Type(="H3")
 4-8  I5 Along-track run-off after 0 hours (meters)
10-14 I5 Cross-track run-off after 0 hours (meters)
16-20 I5 Radial run-off after 0 hours(meters)
22-26 I5 Along-track run-off after 6 hours (meters)
28-32 I5 Cross-track run-off after 6 hours (meters)
34-38 I5 Radial run-off after 6 hours(meters)
40-44 I5 Along-track run-off after 24 hours (meters)
46-50 I5 Cross-track run-off after 24 hours (meters)
52-56 I5 Radial run-off after 24 hours (meters)
******************************************************************************/
void Cpf::parseAccuracyHeader( const std::string& line, Header* header )
{
    yourASSERTION( header );
    std::istringstream s;
    s.str( line );

    if( header->nh3++ > 1 )
        yourTHROW( "Multiple H3 records encountered.");

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "h3" && recType != "H3" )
        yourTHROW( "Error reading cpf record H3." );

    for(long i = 0; i < 3; i++ )
    {
        header->runoff.alongTrack[i]  = convert<double>( s, ok);
        header->runoff.acrossTrack[i] = convert<double>( s, ok);
        header->runoff.radial[i]      = convert<double>( s, ok);
    }

    if( !ok )
    {
        for(long i = 0; i < 3; i++ )
        {
            header->runoff.alongTrack[i]  = 0;
            header->runoff.acrossTrack[i] = 0;
            header->runoff.radial[i]      = 0;
        }
    }

    return;
}

/******************************************************************************
Header type 4 Transponder information
 1-2  A2    Record Type (= "H4")
 4-15 F12.5 Pulse Repetition Frequency (PRF) in Hz
17-26 F10.4 Transponder transmit delay in microseconds
28-38 F11.2 Transponder UTC offset in microseconds
40-50 F11.2 Tranponder Oscillator Drift in parts in 1015
******************************************************************************/
void Cpf::parseTransponderHeader( const std::string& line, Header* header )
{
    yourASSERTION( header );
    if( header->nh4++ > 1 )
        yourTHROW( "Multiple H4 records encountered.");

    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "h4" && recType != "H4" )
        yourTHROW( "Error reading cpf record H4.");

    header->transponder.pulseFrequency  = convert<double>( s, ok ); 
    header->transponder.transmitDelay   = convert<double>( s, ok ); 
    header->transponder.utcOffset       = convert<double>( s, ok ); 
    header->transponder.oscillatorDrift = convert<double>( s, ok ); 

    if( !ok )
        yourTHROW( "Error reading cpf record H4.");

    return;
}

/******************************************************************************
Header type 5 Centre Of Mass information
 1-2  A2    Record Type (= "H5")
 4-10 F7.4  Approximate centre of mass correction in (positive) meters
******************************************************************************/
void Cpf::parseCentreOfMassHeader( const std::string& line, Header* header )
{
    yourASSERTION( header );
    if( header->nh5++ > 1 )
        yourTHROW( "Multiple H5 records encountered.");

    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "h5" && recType != "H5" )
        yourTHROW( "Error reading cpf record H5.");

    header->centreOfMassCorr  = convert<double>( s, ok ); 

    if( !ok )
        yourTHROW( "Error reading cpf record H5.");

    return;
}

/******************************************************************************
Header type 9 End of header (Last header record)
 1-2 A2 Record Type (= "H9")
******************************************************************************/
void Cpf::parseLastHeader( const std::string& line, Header* header )
{
    yourASSERTION( header );
    if( header->nh9++ > 1 )
        yourTHROW( "Multiple H9 records encountered.");

    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok );
    if( !ok || (recType != "h9" && recType != "H9") )
        yourTHROW( "Error reading cpf record H9." );

    header->complete = true;
    return;
}

/******************************************************************************
Record type 10 Position (transmit required for all targets; receive for distant targets 
and transponders. )

 1-2 A2 Record Type (= "10")
 I1 Direction flag (transmit = 1; receive = 2)
I5 Modified Julian Date (MJD)
F13.6 Seconds of Day (UTC) (Transmit or receive)
I2    Leap second flag (= 0 or the value of the new leap second)
F17.3 Geocentric X position in meters
F17.3 Geocentric Y position in meters
F17.3 Geocentric Z position in meters
******************************************************************************/
void Cpf::parsePositionRecord( const std::string& line, const Header& header )
{
    // If last header not detected then the header is incomplete.
    if( !header.complete )
        yourTHROW( " Data record encountered before all headers read.");

    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "10" )
        yourTHROW( "Error reading cpf record 10.");

    long direction  = convert<long>(   s, ok ); 
    long mjd        = convert<long>(   s, ok ); 
    double secs     = convert<double>( s, ok ); 
    long leapSecond = convert<long>(   s, ok ); 
    double x        = convert<double>( s, ok ); 
    double y        = convert<double>( s, ok ); 
    double z        = convert<double>( s, ok ); 

    if( !ok )
        yourTHROW( "Error reading cpf record 10.");

    YourTimeDateClass epoch = YourTimeDateClass::modJulianDate( mjd );
    epoch += YourTimeDateClass::seconds( secs + leapSecond);
    if( direction == 0) 
    {
        m_lightTimeCorrected = false;
        direction = 1;
    }

    if( direction == 1  )
    {
        if( m_biDirectionalPos )
            yourTHROW( "Directional records out of sequence.");
        m_epoch = epoch; 

        // set receive time same as fire or transmit time
        m_epochRx     = m_epoch; 
        m_xPositionTx = x;
        m_yPositionTx = y;
        m_zPositionTx = z;
        m_xPositionRx = -x;
        m_yPositionRx = -y;
        m_zPositionRx = -z;
        m_complete = true;
    }
    else if( direction == 2 )
    {
        m_epochRx = epoch; // receive time
        long sgn = sign( x, y, z, m_xPositionTx, m_yPositionTx, m_zPositionTx );
        m_xPositionRx = sgn*x;
        m_yPositionRx = sgn*y;
        m_zPositionRx = sgn*z;
        m_biDirectionalPos = true;
    }
    else
        yourTHROW( "Error reading cpf record 10.");

    
    return;
}

/******************************************************************************
Record type 20 Velocity (tranmit required for all targets; receive for distant 
targets and transponders)

 1-2 A2 Record Type (= "20")
 I1 Direction flag (transmit = 1; receive = 2)
F19.6 Geocentric X velocity in meters/second
F19.6 Geocentric Y velocity in meters/second
F19.6 Geocentric Z velocity in meters/second

******************************************************************************/
void Cpf::parseVelocityRecord( const std::string& line, const Header& header )
{
    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "20" )
        yourTHROW;

    long direction  = convert<long>( s, ok );   
    double u        = convert<double>( s, ok ); 
    double v        = convert<double>( s, ok ); 
    double w        = convert<double>( s, ok ); 

    if( !ok )
        yourTHROW( "Error reading cpf record 20.");

    if( direction == 0) 
    {
        m_lightTimeCorrected = false;
        direction = 1;
    }
    if( direction == 1 )
    {
        if( m_biDirectionalVel )
            yourTHROW( "Directional records out of sequence.");
        m_xVelocityTx = u;
        m_yVelocityTx = v;
        m_zVelocityTx = w;
        m_xVelocityRx = -u;
        m_yVelocityRx = -v;
        m_zVelocityRx = -w;
    }
    else if( direction == 2 )
    {
        long sgn = sign( u, v, w, m_xVelocityTx, m_yVelocityTx, m_zVelocityTx );
        m_xVelocityRx = sgn*u;
        m_yVelocityRx = sgn*v;
        m_zVelocityRx = sgn*w;
        m_biDirectionalVel = true;
    }
    else
        yourTHROW( "Error reading cpf record 20.");

    m_irvCompatible = header.irvCompatible;
    return;
}

/******************************************************************************
Record type 30 Corrections (All targets computed from a solar system ephemeris)

 1-2 A2 Record Type (= "30")
 A1 Direction flag (transmit = 1; receive = 2)
F18.6 X stellar aberration correction in meters
F18.6 Y stellar aberration correction in meters
F18.6 Z stellar aberration correction in meters
F5.1  Relativistic range correction in nsec (positive number)
******************************************************************************/
void Cpf::parseCorrectionRecord( const std::string& line )
{
    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "30" )
        yourTHROW( "Error reading cpf record 30.");

    long direction = convert<long>( s, ok );   
    double xCor    = convert<double>( s, ok ); 
    double yCor    = convert<double>( s, ok ); 
    double zCor    = convert<double>( s, ok ); 
    double rCor    = convert<double>( s, ok ); 
    if( !ok )
        yourTHROW( "Error reading cpf record 30.");

    if( direction == 0) 
    {
        m_lightTimeCorrected = false;
        direction = 1;
    }
    if( direction == 1 )
    {
        if( m_biDirectionalCor )
            yourTHROW( "Directional records out of sequence.");
        m_xAberrationTx = xCor;
        m_yAberrationTx = yCor;
        m_zAberrationTx = zCor;
        m_xAberrationRx = -xCor;
        m_yAberrationRx = -yCor;
        m_zAberrationRx = -zCor;

        m_relCorrTx     = fabs(rCor);
        m_relCorrRx     = fabs(rCor);
    }
    else if( direction == 2 )
    {
        long sgn = sign( xCor, yCor, zCor, m_xAberrationRx, m_yAberrationRx, m_zAberrationRx );
        m_xAberrationRx = sgn*xCor;
        m_yAberrationRx = sgn*yCor;
        m_zAberrationRx = sgn*zCor;
        m_relCorrRx     = fabs(rCor);
        m_biDirectionalCor = true;
    }
    else
       yourTHROW( "Error reading cpf record 30.");

    return;
}

/******************************************************************************
Record type 40 Transponder specific (Transponders)

 1-2 A1 Record Type (= "40")
F6.3 Oscillator relativity correction in meters/second
******************************************************************************/
void Cpf::parseTransponderRecord( const std::string& line, const Header& header) //, Transponder* transponder)
{
    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok ); 
    if( recType != "40" )
        yourTHROW( "Error reading cpf record 40.");

    if( header.nh4 == 0 )
        yourTHROW("No transponder header prior to a transponder record.");
    m_transponder = header.transponder;

    double cor = convert<double>( s, ok ); 
    if( !ok )
        yourTHROW( "Error reading cpf record 40.");

    m_transponder.oscillatorRelCor = cor; 
    return;
}

/******************************************************************************
Record type 99 Ephemeris Trailer (last record in ephemeris)
 1-2 A2 Record Type (= "99")
******************************************************************************/
void Cpf::parseLastRecord( const std::string& line )
{
    if( m_nlr++ > 1 )
        yourTHROW( "Multiple record 99 encountered.");

    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok );
    if( !ok || recType != "99" )
        yourTHROW( "Error reading cpf record 99.");

    return;
}

/******************************************************************************
******************************************************************************/
void Cpf::addHeaderData( const Header& header )
{
    m_quality        = Element::Quality::okFresh;

    m_objectName     = header.target;
    m_targetId       = header.identifier;
    m_created        = header.created;
    m_supplierId     = header.source;
    m_cospar         = header.cospar; 
    m_norad          = header.norad;
    m_sic            = header.sic;
    m_sequence       = header.sequence;
    m_targetType     = header.targetType;
    m_referenceFrame = header.referenceFrame;
    m_angleType      = header.angleType;
    m_comCorr        = header.centreOfMassCorr;
    m_comApplied     = header.comApplied;
    return;
}


