#if ! defined( _EOP_H_ )
#define _EOP_H_

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "YourTimeDateClass.h"
#include <string>
#include <vector>

// Structure to contain Earth Orientation Parameters (eop).
struct Eop
{
    Eop()
    {
        epoch.setMJD(0,0);
        xPolar = yPolar = 0.0;
        ut1_utc = YourTimeIntervalClass::seconds(0);
    }
    YourTimeDateClass epoch;
    YourTimeIntervalClass ut1_utc;
    double xPolar;
    double yPolar;
};

//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// This class provides a container for a number of earth 
// orientation parameters (EOP) records. Parameter data can be obtained from
// the USNO web site as a SER7 text file or as an IERS Bulletin A file. 
// Parameters so provided include
// 1. polar motion x-coordinate in arcseconds
// 2. polar motion y-coordinate in arcseconds
// 3. time correction UT1 - UTC in seconds
//
// Remarks:
// Linear interpolation between nearest times is performed to calculate 
// parameters at a given epoch. No extrapolation is performed. Hence any
// request to obtain parameters at a time when no data is available covering
// that epoch will fail. That is, an access method will return false or 
// default values are set to zero.
//
// Example: 
//   // Construct an EOP class
//   EOP eop( "c:\directory\eop.txt" );
//
//   // Define an epoch and access the data
//   YourTimeDateClass date;
//   double xp, yp;
//   YourTimeIntervalClass utdelta;
//   bool result = eop.get(date, &xp, &yp, &utdelta);
// or
//   // Access an Eop record
//   Eop r;
//   r = eop.get( date );
//-----------------------------------------------------------------------------
class EOP
{
public:
    // Constructor
    // Argument: filename - full name & path of the EOP source data file.
    EOP();
    EOP( const std::string &filename );
    virtual ~EOP( );

    //  Methods to add a set of EOP parameters
    void add( Eop eop ){ m_aEop.push_back( eop ); }
    void add( YourTimeDateClass epoch, double xPolar, double yPolar, YourTimeIntervalClass ut1_utc);

    // Methods to access EOP parameters
    // Arguments: epoch    -   time of required values.
    //            xPolar   -   x-coordinate of polar motion (radians)
    //            yPolar   -   y-coordinate of polar motion (radians)
    //            ut1_utc  -   UT1-UTC as precision time interval.
    // Return Value: True if valid values available. False otherwise, e.g. if no data available.
    bool get(const YourTimeDateClass& epoch, double& xPolar, double& yPolar,
                                       YourTimeIntervalClass& ut1_utc) const;

    Eop get( const YourTimeDateClass& epoch ) const;
    YourTimeIntervalClass ut1_utc( const YourTimeDateClass& epoch ) const;

private:

    friend class PredictorCPF;

    //  Method to extract eop data from CPF source data
    void addCpfRecord( const std::string& line );

    //  Method to extract parameters from source data
    bool parseRecord( const std::string& line, Eop &eop );

    //  Method to read the usual USNO EOP data files
    bool readFile();

    // Template function to convert a std::stringstream to a basic type.
    // Success indicated by the ok flag.
    template <class T>
    inline T convert( std::istringstream& iss, bool& ok)
    {
        T result;   
        iss.clear();
        if( !(iss >> result ) )
            ok = false;
        return result;
    }

    bool m_fileOk, m_dataOk;
    std::string m_filename;
    std::vector<Eop> m_aEop;

};	   // class EOP

#endif // _EOP_H_
