#include "Track.h"
#include "YourException.h"

/******************************************************************************
 Constructor
******************************************************************************/
OrbitalTrack::OrbitalTrack(  YourTimeDateClass     startPredictions,
                             YourTimeDateClass     stopPredictions,
                             YourTimeIntervalClass stepSize ) : Track( startPredictions, stopPredictions, stepSize )
{
    m_trackType = Track::Type::orbital;
    m_targetResponseDelay = YourTimeDateClass::seconds(0);

}

/******************************************************************************
 Destructor
******************************************************************************/
OrbitalTrack::~OrbitalTrack( )
{
}

/******************************************************************************
Method to add a new node i.e. a set of prediction data at a given point to the
time series of such points. It also sets up a number of arrays for data
associated with each prediction segment. One segment corresponds to predictions
generated from a given set of predictions. i.e. from a CPF file.
These arrays include for each segment
1. start and stop time - time segments may overlap
2. start and stop index 
3. if we have return leg predictions, velocity, etc.
******************************************************************************/
void OrbitalTrack::appendNode( YourTimeDateClass epoch,
                                   bool newSegment, bool biDirectional, bool lightTimeCorrected,
                                   const YourMatrix<double>& posVelCor,
                                   double xPolar, double yPolar, YourTimeIntervalClass ut1_utc )
{
    // If these data are the first of a new segment, generated from a fresh
    // element, then update the time and index arrays. Otherwise, it is
    // a continuation of current segment.

    m_maxIndex++;
    if( newSegment )
    {
        // Create a copy of the element to save in an array.
        m_segmentStartIndex.push_back( m_maxIndex );
        m_segmentStopIndex.push_back ( m_maxIndex );
        m_segmentStartTime.push_back( epoch );
        m_segmentStopTime.push_back ( epoch );
        m_segmentStartTimeBiasPreset.push_back( YourTimeDateClass::seconds(0) );
        m_segmentStopTimeBiasPreset.push_back( YourTimeDateClass::seconds(0) );
        m_segmentTimeBiasLocal.push_back( YourTimeDateClass::seconds(0) );
        m_segmentPosBiasAzim.push_back( 0 );
        m_segmentPosBiasElev.push_back( 0 );
        m_startPredicts = m_segmentStartTime[0];
        m_segmentBiDirectional.push_back( biDirectional );
        m_lightTimeCorrected = lightTimeCorrected;
        m_centreOfMassCorr = posVelCor[6][2];
    }
    m_aXPolar.push_back( xPolar );
    m_aYPolar.push_back( yPolar );
    m_aDUTC.push_back(  ut1_utc );

    m_aX.push_back(        posVelCor[0][0] );
    m_aY.push_back(        posVelCor[0][1] );
    m_aZ.push_back(        posVelCor[0][2] );
    m_arX.push_back(       posVelCor[1][0] );
    m_arY.push_back(       posVelCor[1][1] );
    m_arZ.push_back(       posVelCor[1][2] );
    m_aU.push_back(        posVelCor[2][0] );
    m_aV.push_back(        posVelCor[2][1] );
    m_aW.push_back(        posVelCor[2][2] );
    m_arU.push_back(       posVelCor[3][0] );
    m_arV.push_back(       posVelCor[3][1] );
    m_arW.push_back(       posVelCor[3][2] );
    m_aXAberTx.push_back(  posVelCor[4][0] );
    m_aYAberTx.push_back(  posVelCor[4][1] );
    m_aZAberTx.push_back(  posVelCor[4][2] );
    m_aXAberRx.push_back(  posVelCor[5][0]);
    m_aYAberRx.push_back(  posVelCor[5][1] );
    m_aZAberRx.push_back(  posVelCor[5][2] );
    m_aRelCorTx.push_back( YourTimeDateClass::picoSeconds( posVelCor[6][0] * 1000, 3) );
    m_aRelCorRx.push_back( YourTimeDateClass::picoSeconds( posVelCor[6][1] * 1000, 3) );

    yourASSERTION( !m_segmentStopTime.empty() );
    m_segmentStopTime[m_segmentStopTime.size()-1]  = epoch;
    m_segmentStopIndex[m_segmentStopTime.size()-1] = m_maxIndex;
    m_stopPredicts = epoch;
    m_numOfNodes = m_maxIndex + 1;

    return;
}

/******************************************************************************
 Routine to calculate required x,y,z coordinates given a specific time within 
 the available range of data. 
******************************************************************************/
bool OrbitalTrack::processDataPoint(SiteHandle& site, YourTimeDateClass epoch, long index ) 
{
    // Convert rectangular coordinates to topocentric coordinates for the given site.
    if( !convertITRFToTopocentric(site, epoch, index ) )
        return false;

    site.m_targetRange += m_targetResponseDelay;

    // If the target is below zero degree horizon, then return false;
    normalizeAngles( &site.m_previousAz, &site.m_targetPosSpher[1], &site.m_targetPosSpher[2] );  

    updateTargetCharacteristics( site );

    return true;
}

/******************************************************************************
******************************************************************************/
bool OrbitalTrack::getFlightTime(bool inBound, SiteHandle& site,
                                      YourTimeDateClass epoch, long index,
                                      YourMatrix<double>& targetPos,
                                      YourMatrix<double>& targetVel,
                                      const YourMatrix<double>& sitePos,
                                      const YourMatrix<double>& siteVel,
                                      YourTimeIntervalClass& range,
                                      double* delta )
{
    // If the epoch is outside the available range, then return
    // a delta of zero. This will exit the loop, using the latest avaiable data. 
    if( delta )
        *delta = 0.0;

    if( !getITRFData( site, epoch, index ) )
        return false;

    // Obtain geocenter to target vectors
    for( long i = 0; i < 3; i++)
    {
        targetPos[i][0] = site.m_targetPosRect[inBound?i+3:i];
        targetVel[i][0] = site.m_targetVelRect[inBound?i+3:i];
    }

    double timeGeoc2Target = (1.0e6 * targetPos.norm()- m_centreOfMassCorr)/SpeedOfLight;

    // Obtain site to target vectors
    if( inBound )
    {
        targetPos += sitePos;
        targetVel += siteVel;
    }
    else
    {
        targetPos -= sitePos;
        targetVel -= siteVel;
    }
    double timeSite2Target = (1.0e6 * targetPos.norm() - m_centreOfMassCorr)/SpeedOfLight;

    if( delta)
        *delta = timeGeoc2Target - timeSite2Target;

    range = seconds(timeSite2Target);
    return true;
}

/******************************************************************************
Method to determine topographic position (azimuth, elevation, range etc)
between a station and a target given by CPF records, given the site
location and time.
Hour angle, declination and right ascenscion are also available.

Obtain outbound and inbound range distances and aberration corrections.
******************************************************************************/
bool OrbitalTrack::convertITRFToTopocentric( SiteHandle& site, YourTimeDateClass epoch, long index )
{
    YourMatrix<double> siteVel(3,1);
    YourMatrix<double> sitePos(3,1, site.m_posITRF );
    YourMatrix<double> targetPos(3,1);
    YourMatrix<double> targetVel(3,1);
    YourTimeIntervalClass rangeTx, rangeRx;
    YourTimeDateClass time;
    double dt0, dt = 0;

    // Obtain instantaneous vectors at epoch to provide initial values.
    if( !getFlightTime(false, site, epoch, index, targetPos, targetVel, sitePos, siteVel, rangeTx, &dt ) )
        return false;
    rangeRx = rangeTx;

    if( !m_bestPrecision )
    {
        calculatePointingAngles( false, site, targetPos, targetVel );
    }
    else
    {
        // If supplied data is corrected for light time aberration then iteration is required to find 
        // the transmit time that corresponds to the fire time at the site. The difference between the 
        // distance from target to geocentre and the distance from target to site gives the light-time 
        // delay between the geocenter and site. Hence reduce the ephemeris interpolation time at the 
        // earth's center by this amount, so that the firing time is referred to the time at the telescope.
        if( m_lightTimeCorrected )
        {

            for( long iteration = 0; iteration < 5; iteration++ )
            {
                dt0 = dt;
                time = epoch - YourTimeDateClass::seconds( dt );
                if( !getFlightTime(false, site, time, index, targetPos, targetVel, sitePos, siteVel, rangeTx, &dt ))
                    break;

                // Continue to interpolate with the new time until the change in time is negligible.
                if( fabs(dt-dt0) < 1.0e-10 )
                    break;
            }
            checkOffsets(false, site, targetPos, targetVel, rangeTx);
            calculatePointingAngles( false, site, targetPos, targetVel );

            // For the inbound or receive leg, check if there is any delay associated with a return signal.
            // For example, transponders may introduce an effective delay. Adjust the bounce time by the delay.
            // Since the interpolation time (the bounce time at the target) is already known from 
            // the outbound leg computation, then no iteration is required.
            time += m_targetResponseDelay;
            if( !getFlightTime(true, site, time, index, targetPos, targetVel, sitePos, siteVel, rangeRx ))
                return false;
            checkOffsets(true, site, targetPos, targetVel, rangeRx);
            calculatePointingAngles(true, site, targetPos, targetVel );
        }

        // If the predictions are not corrected for light time aberration, then determine time of impact
        // at target by iteration since the target will move during time of flight.
        if( !m_lightTimeCorrected )
        {
            for( long iteration = 0; iteration < 2; iteration++ )
            {
                time = epoch + rangeTx;
                if( !getFlightTime(false, site, time, index, targetPos, targetVel, sitePos, siteVel, rangeTx ))
                    return false;
            }
            checkOffsets(false, site, targetPos, targetVel, rangeTx);
            calculatePointingAngles( false, site, targetPos, targetVel );

            // The inbound or receive leg must also be corrected for light time aberration since
            // the receiver will have moved during time of flight due to earth rotation. 
            YourMatrix<double> newSite = sitePos;
            for( long iteration = 0; iteration < 2; iteration++ )
            {
                time = epoch + m_targetResponseDelay + rangeTx;
                YourMatrix<double> r3(YourMatrix::rotationMatrix3<double>( YourMatrix::EarthRotationRate * YourTimeDateClass::seconds( rangeTx + m_targetResponseDelay + rangeRx) ) );
                newSite = r3 * sitePos;
                if( !getFlightTime(true, site, time, index, targetPos, targetVel, newSite, siteVel, rangeRx) )
                    return false;
            }
            checkOffsets(true, site, targetPos, targetVel, rangeRx);
            calculatePointingAngles(true, site, targetPos, targetVel );
        }
    }

    // Find total flight time for basic satellite type targets.
    site.m_targetRange = rangeTx + rangeRx;

    // Calculate the magnitude of the velocity component on the Az-El plane.
    double azVelOnSky = site.m_targetVelSpher[1] * cos( site.m_targetPosSpher[2] );
    site.m_targetVelocityAzEl = sqrt( site.m_targetVelSpher[2] * site.m_targetVelSpher[2] + azVelOnSky * azVelOnSky );

    for(long i = 0; i < 3; i++ )
    {
        site.m_targetMeanSpher[i]   = 0.5*(site.m_targetPosSpher[i] + site.m_targetPosSpher[i+3]); 
        site.m_targetMeanSpher[i+3] = 0.5*(site.m_targetVelSpher[i] + site.m_targetVelSpher[i+3]); 
    }

    return true;
}

/******************************************************************************
 Routine to interpolate required data at a given specific time.
******************************************************************************/
bool OrbitalTrack::interpolateData(SiteHandle& site, YourTimeDateClass epoch ) 
{
    site.m_targetPosRect[0] = lagrangianInterpolation( m_aX );
    site.m_targetPosRect[1] = lagrangianInterpolation( m_aY );
    site.m_targetPosRect[2] = lagrangianInterpolation( m_aZ );
    site.m_targetVelRect[0] = lagrangianInterpolation( m_aU );
    site.m_targetVelRect[1] = lagrangianInterpolation( m_aV );
    site.m_targetVelRect[2] = lagrangianInterpolation( m_aW );
    site.m_aberCor[0]       = lagrangianInterpolation( m_aXAberTx );
    site.m_aberCor[1]       = lagrangianInterpolation( m_aYAberTx );
    site.m_aberCor[2]       = lagrangianInterpolation( m_aZAberTx );
    site.m_relCor[0]        = lagrangianInterpolation( m_aRelCorTx );

    if( m_bestPrecision )
    {
        site.m_xPolar           = lagrangianInterpolation( m_aXPolar );
        site.m_yPolar           = lagrangianInterpolation( m_aYPolar );
        site.m_ut1_utc          = lagrangianInterpolation( m_aDUTC );
        site.m_targetPosRect[3] = lagrangianInterpolation( m_arX );
        site.m_targetPosRect[4] = lagrangianInterpolation( m_arY );
        site.m_targetPosRect[5] = lagrangianInterpolation( m_arZ );
        site.m_targetVelRect[3] = lagrangianInterpolation( m_arU );
        site.m_targetVelRect[4] = lagrangianInterpolation( m_arV );
        site.m_targetVelRect[5] = lagrangianInterpolation( m_arW );
        site.m_aberCor[3]       = lagrangianInterpolation( m_aXAberRx );
        site.m_aberCor[4]       = lagrangianInterpolation( m_aYAberRx );
        site.m_aberCor[5]       = lagrangianInterpolation( m_aZAberRx );
        site.m_relCor[1]        = lagrangianInterpolation( m_aRelCorRx );
    }

    return true;
}

/******************************************************************************
 Routine to extract node values given a node index.
******************************************************************************/
bool OrbitalTrack::extractNodeData(SiteHandle& site, long index ) 
{
    if( index < 0 || index >= long(m_aX.size()) )
        return false;

    site.m_index = index;
    site.m_epoch = getEpoch( index, &site.m_segment );
    site.m_targetPosRect[0] =  m_aX[index];
    site.m_targetPosRect[1] =  m_aY[index];
    site.m_targetPosRect[2] =  m_aZ[index];
    site.m_targetVelRect[0] =  m_aU[index];
    site.m_targetVelRect[1] =  m_aV[index];
    site.m_targetVelRect[2] =  m_aW[index];
    site.m_aberCor[0]       =  m_aXAberTx[index];
    site.m_aberCor[1]       =  m_aYAberTx[index];
    site.m_aberCor[2]       =  m_aZAberTx[index];
    site.m_relCor[0]        =  m_aRelCorTx[index];

    if( m_bestPrecision )
    {
        site.m_xPolar           =  m_aXPolar[index];
        site.m_yPolar           =  m_aYPolar[index];
        site.m_ut1_utc          =  m_aDUTC[index];
        site.m_targetPosRect[3] =  m_arX[index];
        site.m_targetPosRect[4] =  m_arY[index];
        site.m_targetPosRect[5] =  m_arZ[index];
        site.m_targetVelRect[3] =  m_arU[index];
        site.m_targetVelRect[4] =  m_arV[index];
        site.m_targetVelRect[5] =  m_arW[index];
        site.m_aberCor[3]       =  m_aXAberRx[index];
        site.m_aberCor[4]       =  m_aYAberRx[index];
        site.m_aberCor[5]       =  m_aZAberRx[index];
        site.m_relCor[1]        =  m_aRelCorRx[index];
    }

    return true;
}
/******************************************************************************
******************************************************************************/
void OrbitalTrack::calculatePointingAngles(bool inBound, SiteHandle& site,
                                                YourMatrix<double>& targetPos,
                                                YourMatrix<double>& targetVel )
{
    // Apply stellar aberration corrections
    YourMatrix<double> aberration(3,1, &site.m_aberCor[inBound?3:0] );

    if( inBound )
    {
        targetPos  *= -1;
        targetVel  *= -1;
        aberration *= -1;
    }
    targetPos += aberration;

    // Apply rotations to convert to topocentric equatorial and horizontal
    YourMatrix<double> posTEQ( rotationMatrix3<double>( site.m_longitude )* targetPos  );
    YourMatrix<double> velTEQ( rotationMatrix3<double>( site.m_longitude )* targetVel );
    YourMatrix<double> posSEZ( rotationMatrix2<double>( PI_ON_2 - site.m_geodeticLatitude ) * posTEQ );
    YourMatrix<double> velSEZ( rotationMatrix2<double>( PI_ON_2 - site.m_geodeticLatitude ) * velTEQ );

    // Apply rectangular to spherical transformations
    convertToSpherical( posTEQ, velTEQ );
    convertToSpherical( posSEZ, velSEZ );
    
    targetPos = posSEZ;
    targetVel = velSEZ;

    // Change seond entry to be azimuth values positive CW from northing rather than
    // longitude, which it positive CCW from southing
    targetPos[1][0] = PI - targetPos[1][0];
    targetVel[1][0] = - targetVel[1][0];

    site.m_targetRA  = posTEQ[1][0] * HOUR_PER_RAD;
    site.m_targetDec = posTEQ[2][0] * DEG_PER_RAD;

    for(long i = 0; i < 3; i++ )
    {
        site.m_targetPosSpher[i+3] = site.m_targetPosSpher[inBound?i+3:i]= targetPos[i][0];
        site.m_targetVelSpher[i+3] = site.m_targetVelSpher[inBound?i+3:i]= targetVel[i][0];
    }
    return;
}
