#if ! defined( _ROTANGLE_H_ )
#define _ROTANGLE_H_

#include "YourTimeDateClass.h"
#include <string>
#include <vector>

//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// This class provides a set of rotation angle and Greenwich Apparent Sidereal
// Time (gast) data, primaruily for use with CPF prediction file data sets.
//
// Remarks:
// Linear interpolation between nearest times is performed to calculate 
// parameters at a given epoch. No extrapolation is performed. Hence any
// request to obtain parameters at a time when no data is available covering
// that epoch will fail (i.e. the access method will return false).
//-----------------------------------------------------------------------------
class RotationAngles
{
public:

    //  Constructor
    RotationAngles();
    virtual ~RotationAngles();

    enum Type
    {
        unknownAngleType = 0,
        lunarEular,             // Lunar Euler angles: 
        raDecMeridian           // North pole Right Ascension and Declination, and angle to prime meridian (?0, ?0, and W)
    };

    struct Record
    {
        Record()
        {
            epoch.setMJD(0,0);
            gast = angle1 = angle2 = angle3 = 0.0;
        }
        YourTimeDateClass epoch;
        double angle1;
        double angle2;
        double angle3;
        double gast;
    };

    //  Method to set the rotation angle type.
    void rotationAngleTypeIs( Type type ){ m_angleType = type;};

    //  Method to access Rotation Angle parameters
    // Arguments: epoch - time of required values.
    // angle1   -  rotation angle 1 (degrees) 
    // angle2   -  rotation angle 2 (degrees) 
    // angle3   -  rotation angle 3 (degrees) 
    // gast     -  Greenwich apparent sidereal time (hours)
    // Return Value: True if valid values available. False otherwise, e.g. if no data available.
    bool get(YourTimeDateClass epoch, double& angle1, double& angle2, double& angle3, double& gast )const;

    //  Method to access Greenwich Apparent Sidereal Time
    // Arguments: epoch  -  epoch of required correction.
    // Return Value: Value of gast, or zero, if no data is available.
    double gast( YourTimeDateClass epoch ) const;

    //  Method to access the rotation angle type.
   Type rotationAngleType() const {return m_angleType;}; 

private:

    friend class PredictorCPF;

    //  Method to extract parameters from source data
    void addCpfRecord( const std::string& line );

    // Template function to convert a std::stringstream to a basic type.
    // Success indicated by the ok flag.
    template <class T>
    inline T convert( std::istringstream& iss, bool& ok)
    {
        T result;   
        iss.clear();
        if( !(iss >> result ) )
            ok = false;
        return result;
    }

    Type m_angleType;
    std::vector<Record> m_aAngles;

};  // class RotationAngles

#endif // _ROTANGLE_H_
